/*
** cgi.c
**
** Copyright (c) 1994-1995 Peter Eriksson <pen@signum.se>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <alloca.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <wait.h>
#include <errno.h>
#include <signal.h>
#include <syslog.h>

#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/types.h>

#include "phttpd.h"

#define MAXENV	256

static char *cgi_user = NULL;
static char *cgi_group = NULL;
static char *cgi_path = NULL;
static char *copy_environ = NULL;

static int cgi_uid = 60001; /* Nobody */
static int cgi_gid = 60001; /* Nobody */
static int user_cgi = 1;
static int max_environ = MAXENV;
static int max_processes = 0;

static sema_t proc_sema;



static char *cgi_setenv(char **envp, const char *var, const char *val)
{
    char *buf;
    int i, len;


    if (debug > 5)
	fprintf(stderr, "cgi_setenv: var=%s, val=%s\n",
		var, val ? val : "<null>");
    
    if (val == NULL)
	return NULL;
    
    buf = s_malloc(strlen(var)+strlen(val)+2);
    
    strcpy(buf, var);
    strcat(buf, "=");
    
    len = strlen(buf);
    
    strcat(buf, val);
    
    for (i = 0; i < max_environ && envp[i]; i++)
	if (strncmp(envp[i], buf, len) == 0)
	    break;

    if (i >= max_environ)
	return NULL;

    if (envp[i])
	s_free(envp[i]);
    
    envp[i] = buf;
    return buf;
}


static char *cgi_setenv_i(char **envp, const char *var, int num)
{
    char buf[256];

    sprintf(buf, "%d", num);
    return cgi_setenv(envp, var, buf);
}


static int cse_header(hashentry_t *hep, void *envp)
{
    char *name, *cp;

    
    if (strcmp(hep->key, "AUTHORIZATION") == 0)
	return 0;
    
    name = (char *) alloca(strlen(hep->key)+6);
    strcpy(name, "HTTP_");
    cp = name+5;
    strcpy(cp, hep->key);
    while (*cp)
    {
	if (*cp == '-')
	    *cp = '_';
	++cp;
    }

    cp = cgi_setenv(envp, name, hep->data);
    
    return 0;
}


static char **cgi_setup_env(const char *method,
			    char *url,
			    struct connectioninfo *cip,
			    struct httpinfo *hip,
			    struct mimeinfo *mip,
			    const char *path,
			    char *path_info,
			    char *path_translated)
{
    char **envp, *buf, *request;

    
    envp = s_malloc((max_environ+1) * sizeof(char *));
    if (envp == NULL)
	return NULL;

    cgi_setenv(envp, "DOCUMENT_ROOT", server_home);
    cgi_setenv(envp, "TZ", tzname[0]);
    cgi_setenv(envp, "PATH", path);

    buf = (char *) alloca(strlen(server_version) + 8);
    sprintf(buf, "phttpd/%s", server_version);
    
    cgi_setenv(envp, "SERVER_SOFTWARE", buf);
    cgi_setenv(envp, "SERVER_NAME", server_host);
    cgi_setenv(envp, "GATEWAY_INTERFACE", "CGI/1.1");
    cgi_setenv(envp, "REQUEST_METHOD", method);
    cgi_setenv(envp, "PATH_INFO", path_info);
    
    cgi_setenv_i(envp, "SERVER_PORT", server_port);
    
    if (path_translated && path_translated[0])
    {
	if (path_translated[0] != '/')
	{
	    buf = (char *) alloca(strlen(server_home) + 
				  strlen(path_translated) + 2);
	    strcpy(buf, server_home);
	    if (strcmp(path_translated, ".") != 0)
	    {
		strcat(buf, "/");
		strcat(buf, path_translated);
	    }
	    
	    cgi_setenv(envp, "PATH_TRANSLATED", buf);
	}
	else
	    cgi_setenv(envp, "PATH_TRANSLATED", path_translated);
    }

    if (cip)
    {
	cgi_setenv(envp, "LOCAL_ADDR",  cip->server.addr_s);
	cgi_setenv(envp, "LOCAL_HOST", get_hostname(&cip->server));
	
	cgi_setenv(envp, "REMOTE_ADDR", cip->client.addr_s);
	cgi_setenv(envp, "REMOTE_HOST", get_hostname(&cip->client));
	cgi_setenv(envp, "REMOTE_IDENT", get_ident(&cip->client));
	
	cgi_setenv(envp, "AUTH_TYPE", cip->auth.type);
	cgi_setenv(envp, "REMOTE_USER", cip->auth.user);
    }

    if (hip)
    {
	cgi_setenv(envp, "SERVER_PROTOCOL", hip->version);
	cgi_setenv(envp, "SCRIPT_NAME", url);
	cgi_setenv(envp, "QUERY_STRING", hip->request);
    }
	    
    if (mip)
    {
	cgi_setenv(envp, "CONTENT_TYPE",
		   mime_getheader(mip, "CONTENT-TYPE", 1));

	cgi_setenv(envp, "CONTENT_LENGTH",
		   mime_getheader(mip, "CONTENT-LENGTH", 1));

	ht_foreach(&mip->table, cse_header, (void *) envp);
    }



    cgi_setenv(envp, "PHTTPD_ORIG_URL", hip->orig_url);
    cgi_setenv(envp, "PHTTPD_ORIG_REQUEST", hip->orig_request);
    
    /* Non CGI/1.1 conformant extension, but very nice to have... */
    if (hip->request && hip->request[0])
    {
	char **argv;
	
	request = (char *) alloca(strlen(hip->request)+1);
	strcpy(request, hip->request);
	
	if ((argv = strsplit(request, '&', 0)) != NULL)
	{
	    int i;
	    
	    for (i = 0; argv[i]; i++)
	    {
		char **varval, *var_name;

		
		if ((varval = strsplit(argv[i], '=', 2)) != NULL)
		{
		    if (debug > 4)
			fprintf(stderr, "varval[0] = %s, varval[1] = %s\n",
				varval[0],
				varval[1]);
		    
		    if (varval[0] && varval[1])
		    {
			var_name = (char *) alloca(strlen(varval[0])+14);
			strcpy(var_name, "PHTTPD_QUERY_");
			strcat(var_name, varval[0]);
			
			cgi_setenv(envp, var_name, varval[1]);
		    }

		    s_free(varval);
		}
	    }

	    s_free(argv);
	}
    }

    if (copy_environ)
    {
	char *var, *cp, *cea, *val;

	cea = alloca(strlen(copy_environ)+1);
	strcpy(cea, copy_environ);

	var = strtok_r(cea, ":", &cp);
	while (var)
	{
	    val = getenv(var);

	    cgi_setenv(envp, var, val);
	    
	    var = strtok_r(NULL, ":", &cp);
	}
	
    }
    
    return envp;
}


static void cgi_free_env(char **envp)
{
    int i;

    
    if (envp == NULL)
	return;

    for (i = 0; i < max_environ && envp[i]; i++)
	s_free(envp[i]);

    s_free(envp);
}

static struct options cgi_cfg_table[] = 
{
    { "default-user", 	T_STRING, &cgi_user,   NULL },
    { "default-group", 	T_STRING, &cgi_group,   NULL },
    { "path",		T_STRING, &cgi_path,  NULL },
    { "copy-environ",   T_STRING, &copy_environ, NULL },
    
    { "allow-user-cgi", T_BOOL,   &user_cgi,  NULL },

    { "max-processes",  T_NUMBER, &max_processes, NULL },
    { "max-environ",    T_NUMBER, &max_environ, NULL  },
    
    { NULL, -1, NULL, NULL }
};


int pm_init(const char **argv)
{
    struct passwd *pwp, pwb;
    char pbuf[1024];
    char *cfg_path, *cp;
    const char *name;


    name = argv[0];
    
    if (debug > 1)
	fprintf(stderr, "*** cgi/pm_init(\"%s\") called ***\n", name);

    cfg_path = s_malloc(strlen(name)+6);
    strcpy(cfg_path, name);
    
    cp = strrchr(cfg_path, '.');
    if (cp && strcmp(cp, ".so") == 0)
	*cp = '\0';
    
    strcat(cfg_path, ".conf");
    if (config_parse_file(cfg_path, cgi_cfg_table, 0) < 0)
    {
	if (debug > 1)
	    fprintf(stderr, "config_parse_file() failed\n");
	
	return -1;
    }
    
    if (config_parse_argv(argv+1, cgi_cfg_table) < 0)
    {
	if (debug > 1)
	    fprintf(stderr, "config_parse_file() failed\n");
	return -1;
    }
    
    if (uidgid_get(cgi_user, cgi_group,
		   &cgi_uid, &cgi_gid,
		   &pwp, &pwb, pbuf, sizeof(pbuf)) < 0)
    {
	syslog(LOG_ERR,
	       "cgi: uidgid_get(\"%s\", \"%s\") failed, using default",
	       cgi_user, cgi_group);
    }
    
    if (cgi_uid == -1)
        cgi_uid = server_uid;
    
    if (cgi_gid == -1)
        cgi_gid = server_gid;
    
    if (cgi_path == NULL)
	cgi_path = s_strdup("/bin:/usr/bin");

    if (max_processes > 0)
	sema_init(&proc_sema, max_processes, USYNC_THREAD, NULL);
    
    return 0;
}


void pm_exit(void)
{
    if (debug > 1)
	fprintf(stderr, "*** cgi/pm_exit() called ***\n");

    if (cgi_user)
	s_free(cgi_user);
    
    if (cgi_group)
	s_free(cgi_group);
    
    if (cgi_path)
	s_free(cgi_path);
}


struct relayinfo
{
    int read_fd;
    int write_fd;
    int pid;
    int content_length;
};


static void *cgi_relay_data(void *misc)
{
    struct relayinfo *rip = (struct relayinfo *) misc;

    fd_nrelay(rip->read_fd, rip->write_fd, 0, rip->content_length);

    if (debug > 4)
	fprintf(stderr, "CGI: Closed Client->CGI descriptor\n");
    
    fd_close(rip->write_fd);
    
    return NULL;
}


int pm_request(struct connectioninfo *cip)
{
    int pid, status, len, why, code;
    int content_length = -1;
    int headers_sent = 0;
    int lineno = 0;
    int p2c_pipe[2];
    int c2p_pipe[2];
    int flags;
    int result;
    char **nargv = NULL;
    char buf[1024];
    char *cp, *rp;
    char **envp;
    char *wdir = NULL;
    int uid = -1, gid = -1;
    struct stat sb;
    char path[1025];
    char path_translated[1025];
    struct relayinfo rb;
    thread_t tid;
    char *url;
    char *path_info;
    int fd = cip->fd;
    struct httpinfo *hip = cip->hip;


    c2p_pipe[0] = c2p_pipe[1] = -1;

    if (debug > 1)
    {
	fprintf(stderr, "*** cgi/pm_request(\"%s\") called ***\n",
		hip->url);
	if (hip->request)
	    fprintf(stderr, "\thip->request = \"%s\"\n", hip->request);
    }

    result = 0;

    len = strlen(hip->url);
    url = (char *) alloca(len + 1);
    strcpy(url, hip->url);

    cp = NULL;
    
    while ((rp = url_expand(url, path, sizeof(path), &sb, &uid, &gid)) == NULL)
    {
	/* Do some CGI special handling to support the use of things
	 * like:
         *
	 *  	/cgi-bin/imagemap/foo/bar/fubar.gif
         *
	 *  where /cgi-bin/imagemap is the CGI program...
	 *
	 * Should perhaps be modified to try things from the beginning
	 * first instead of backwards. Later.
	 */
	cp = strrchr(url, '/');
	if (cp == NULL)
	    break;
	
	*cp = '\0';
    }

    if (cp == NULL && len > 0)
	if (url[len-1] == '/')
	{
	    cp = url+len-1;
	    *cp++ = 0;
	}
    
    /* We don't allow users to run CGI scripts? */
    if (user_cgi == 0 && uid != -1)
        return -1;

    if (rp == NULL || !S_ISREG(sb.st_mode))
	return -1;

    if (cp)
    {
	path_info = hip->url + (cp - url);
	if (*path_info == 0)
	    path_info = "/";

	path_translated[0] = '\0';
	url_expand(path_info,
		   path_translated, sizeof(path_translated),
		   NULL, NULL, NULL);
    }
    else
    {
	path_info = NULL;
	path_translated[0] = '\0';
    }

    status = 0;
    envp = cgi_setup_env(hip->method,
			 url,
			 cip, hip, hip->mip,
			 cgi_path,
			 path_info,
			 path_translated);

    if (hip->mip)
    {
	cp = mime_getheader(hip->mip, "CONTENT-LENGTH", 1);
	if (cp)
	    content_length = atoi(cp);
    }
    
    len = strlen(hip->url);

    if (pipe(p2c_pipe) < 0)
    {
	cgi_free_env(envp);
	return error_system(cip, "pipe()");
    }
    
    flags = fcntl(p2c_pipe[0], F_GETFL, 0);
    flags &= ~O_ACCMODE;
    flags |= O_RDONLY;
    fcntl(p2c_pipe[0], F_SETFL, flags);
    
    flags = fcntl(p2c_pipe[1], F_GETFL, 0);
    flags &= ~O_ACCMODE;
    flags |= O_WRONLY;
    fcntl(p2c_pipe[1], F_SETFL, flags);

    fd_reopen(p2c_pipe[1], O_WRONLY, FDT_PIPE);

    if (pipe(c2p_pipe) < 0)
    {
	cgi_free_env(envp);
	return error_system(cip, "pipe()");
    }
    
    flags = fcntl(c2p_pipe[0], F_GETFL, 0);
    flags &= ~O_ACCMODE;
    flags |= O_RDONLY;
    fcntl(c2p_pipe[0], F_SETFL, flags);
    
    flags = fcntl(c2p_pipe[1], F_GETFL, 0);
    flags &= ~O_ACCMODE;
    flags |= O_WRONLY;
    fcntl(c2p_pipe[1], F_SETFL, flags);


    /* Set up the process argv[] array */
    if (hip->request == NULL || strchr(hip->request, '='))
    {
	nargv = (char **) alloca(2 * sizeof(char *));
	if (nargv == NULL)
	    abort();
	
	nargv[0] = hip->url;
	nargv[1] = NULL;
    }
    else
    {
	char *cp, *start;
	int i;
	
	
	i = 3; /* argv[0] + first arg + NULL */
	for (cp = hip->request; *cp; cp++)
	    if (*cp == '+')
		++i;
	
	nargv = (char **) alloca(i * sizeof(char *));
	if (nargv == NULL)
	    abort();
	
	nargv[0] = hip->url;
	i = 1;
	start = hip->request;
	do
	{
	    nargv[i++] = start;
	    
	    cp = strchr(start, '+');
	    if (cp)
	    {
		*cp++ = '\0';
		start = cp;
	    }
	} while (cp);
	nargv[i] = NULL;
    }

    if (uid == -1)
	uid = cgi_uid;
    if (gid == -1)
	gid = cgi_gid;

    if (max_processes > 0)
	sema_wait(&proc_sema);
    
    pid = proc_run(path,
		   uid, gid,
		   p2c_pipe[0], c2p_pipe[1], 2,
		   nargv, envp,
		   wdir);

    if (pid < 0)
    {
	s_close(p2c_pipe[0]);
        fd_close(p2c_pipe[1]);
	if (c2p_pipe[0] != -1)
	    s_close(c2p_pipe[0]);
	if (c2p_pipe[1] != -1)
	    s_close(c2p_pipe[1]);

	if (max_processes > 0)
	    sema_post(&proc_sema);
	
	return error_system(cip, "fork1()");
    }

    s_close(p2c_pipe[0]);

    rb.read_fd = fd;
    rb.write_fd = p2c_pipe[1];
    rb.pid = pid;
    rb.content_length = content_length;
    
    if (thr_create(NULL,
		   stack_size, 
		   (void *(*)(void *)) cgi_relay_data,
		   (void *) &rb,
		   bound_threads ? THR_BOUND : 0,
		   &tid))
    {
        fd_close(p2c_pipe[1]);

	if (c2p_pipe[0] != -1)
	    s_close(c2p_pipe[0]);
	if (c2p_pipe[1] != -1)
	    s_close(c2p_pipe[1]);

	if (max_processes > 0)
	    sema_post(&proc_sema);
	
	return -1;
    }

    s_close(c2p_pipe[1]);
    fd_reopen(c2p_pipe[0], O_RDONLY, FDT_PIPE);

    while (fd_gets(buf, sizeof(buf), c2p_pipe[0]))
    {
	lineno++;
	
	if (*buf == '\r' || *buf == '\n')
	    break;
	
	/* This check is not according to the CGI-standard! */
	if (lineno == 1 && strncmp(buf, "HTTP/", 5) == 0)
	{
	    fd_puts(buf, fd);
	    fd_relay(c2p_pipe[0], fd, 0);
	    buf[0] = '\0';
	    headers_sent = 1;
	    lineno = 0;
	    break;
	}
	
	cp = strchr(buf, ':');
	if (cp == NULL)
	    break;
	
	*cp = '\0';
	
	if (strcasecmp(buf, "LOCATION") == 0)
	{
	    result = 302;
	    http_sendheaders(fd, cip, result, NULL);
	    headers_sent = 1;
	    
	    fd_printf(fd, "Location: %s\n", cp+1);
	}
	else if (strcasecmp(buf, "STATUS") == 0)
	{
	    char buf2[1024], *emsg;
	    int rval;
	    
	    
	    emsg = NULL;
	    
	    rval = sscanf(cp + 1, " %d %[^\n\r]", &result, buf2);
	    if (rval < 1)
	    {
		result = http_error(cip, 500,
    "Invalid response from CGI script: Too few arguments to 'STATUS:'");
		headers_sent = 1;
	    }
	    else
	    {
		if (rval == 2)
		    emsg = buf2;
		
		http_sendheaders(fd, cip, result, emsg);
		headers_sent = 1;
	    }
	} else
	{
	    if (headers_sent == 0)
	    {
		result = 200;
		http_sendheaders(fd, cip, result, NULL);
		headers_sent = 1;
	    }
	    
	    *cp = ':';
	    fd_puts(buf, fd);
	}
	
	buf[0] = '\0';
    }
    
    if (lineno != 0)
    {
	int len;
	
	
	if (!headers_sent)
	{
	    result = 200;
	    http_sendheaders(fd, cip, result, NULL);
	    fd_puts("Content-Type: text/plain\n", fd);
	}
	
	if (cip->hip->mip)
	    fd_putc('\n', fd);
	
	len = fd_written(fd);
	
	if (buf[0] && buf[0] != '\n' && buf[0] != '\r')
	    fd_puts(buf, fd);
	
	fd_relay(c2p_pipe[0], fd, 0);
    }
    
    fd_close(c2p_pipe[0]);
    c2p_pipe[0] = -1;

    errno = 0;

    why = proc_wait(pid, &code);
    switch (why)
    {
      case PROC_EXIT:
	if (code < 0)
	    result = -code;
	else if (code > 0)
	    result = http_error(cip,
				500,
				"CGI process exited with exit code %d",
				code);
	break;

      case PROC_SIGNAL:
	result = http_error(cip, 500,
			    "CGI process exited due to signal %d",
			    code);
	break;

      default:
	result = http_error(cip,
			    500,
			    "CGI process terminated unexpectedly");
    }

    cgi_free_env(envp);

    if (debug > 5)
	fprintf(stderr, "before thr_join\n");
    
    thr_join(tid, NULL, NULL);
    if (debug > 5)
	fprintf(stderr, "before fd_close\n");
    
    if (c2p_pipe[0] != -1)
	fd_close(c2p_pipe[0]);

    if (debug > 5)
	fprintf(stderr, "before sema_post\n");
    
    if (max_processes > 0)
	sema_post(&proc_sema);
	
    return result;
}
