/*
** config.c
**
** This is the main file of the config module.
**
** Copyright (c) 1995 Peter Eriksson <pen@signum.se>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <thread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include "phttpd.h"


int pm_init(const char **argv)
{
    const char *name = argv[0];
    
    if (debug > 1)
	fprintf(stderr, "*** libconfig/pm_init(\"%s\") called ***\n", name);

    return 0;
}


void pm_exit(void)
{
    if (debug > 1)
	fprintf(stderr, "*** libconfig/pm_exit() called ***\n");
}


struct miscinfo
{
    int fd;
    char *url;
};

static int print_connection(struct connectioninfo *cip, void *misc)
{
    char buf[256];
    struct miscinfo *mp;



    mp = (struct miscinfo *) misc;

    fd_printf(mp->fd, "<DT>#%d: ", cip->request_no);

    fd_printf(mp->fd, "<DD><A HREF=\"%s?descriptors#%d\">fd=%d</A>",
	      mp->url, (cip->fd < 0 ? -(cip->fd) : cip->fd), cip->fd);
    
    fd_printf(mp->fd, ", tid=%d, client=%s:%d<BR>at=%s<BR>",
	      (int) cip->tid,
	      cip->client.addr_s,
	      cip->client.port,
	      ctime_r(&cip->cn_time, buf, sizeof(buf)));

    if (cip->hip && cip->hip->url)
	fd_printf(mp->fd, "url=%s\n", cip->hip->url);
    
    return 0;
}

static int print_module(struct modinfo *mp, void *misc)
{
    int fd;
    char buf[256];
    
    
    fd = * (int *) misc;

    fd_printf(fd, "%s, use=%d, last modified=%s<BR>\n",
	      mp->name,
	      mp->use,
	      ctime_r(&mp->mtime, buf, sizeof(buf)));
    
    return 0;
}

static int print_fd(int fd, struct fd_buffer *bp, void *misc)
{
    int out_fd;


    out_fd = * (int *) misc;

    if (fd == out_fd)
	return 0;
    
    fd_printf(out_fd, "<A NAME=\"%d\">\n", fd);
    
    if (fd == out_fd)
	fd_printf(out_fd, "<DT><B>%d</B> (this connection)\n", fd);
    else
	fd_printf(out_fd, "<DT><B>%d</B>\n", fd);
    
    fd_printf(out_fd,
	      "<DD>Input: size=%d, state=%d<BR>\n",
	      bp->in.end - bp->in.start,
	      bp->in.state);
    
    fd_printf(out_fd, "Output: size=%d, state=%d\n",
	      bp->out.end - bp->out.start,
	      bp->out.state);

    fd_puts("</A>\n", out_fd);
    return 0;
}


static int print_table_pair(void *value, void *misc)
{
    char **pair = (char **) value;
    int fd = * (int *) misc;

    
    fd_printf(fd, "%-20s %s\n", pair[0], pair[1]);

    return 0;
}

static void print_stats(int fd, cachestat_t *csp, char *label)
{
    fd_printf(fd, "\t<TR><TD>%-8s<TD ALIGN=RIGHT>%10lu<TD ALIGN=RIGHT>%10lu<TD ALIGN=RIGHT>%10lu<TD ALIGN=RIGHT>%10lu\n",
	      label,
	      csp->lookups,
	      csp->reloads,
	      csp->updates,
	      csp->hits);
}

static int http_get_head(struct connectioninfo *cip)
{
    int result, len;
    int fd = cip->fd;
    struct httpinfo *hip = cip->hip;
    

    result = 200;
    
    if (hip->mip != NULL)
    {
	http_sendheaders(fd, cip, result, NULL);

	fd_puts("Pragma: No-Cache\n", fd);
	
	http_sendlastmodified(fd, cip->cn_time);
	
	fd_puts("Content-Type: text/html\n", fd);
    }

    if (strcasecmp(hip->method, "HEAD") == 0)
	return result;

    if (hip->mip != NULL)
	fd_putc('\n', fd);
    
    len = fd_written(fd);

    if (hip->request == NULL)
    {
	html_sysheader(fd, "H2", "Phttpd Configuration Menu");

	fd_puts("<OL>\n", fd);
	
	fd_printf(fd,
		  "<LI><A HREF=\"%s?configuration\">Configuration</A>\n",
		  hip->url);
	
	fd_printf(fd,
		  "<LI><A HREF=\"%s?statistics\">Server Statistics</A>\n",
		  hip->url);
	
	fd_printf(fd,
		  "<LI><A HREF=\"%s?connections\">Current Connections</A>\n",
		  hip->url);

	fd_printf(fd,
		  "<LI><A HREF=\"%s?modules\">Loaded Modules</A>\n",
		  hip->url);
	
	fd_printf(fd,
		  "<LI><A HREF=\"%s?descriptors\">Open Descriptors</A>\n",
		  hip->url);

	if (cntfile_path)
	{
	    fd_printf(fd,
		      "<LI><A HREF=\"%s?requests\">Request Summary</A>\n",
		      hip->url);
	}
	
	fd_puts("</OL>\n", fd);
    }
    
    else if (strcmp(hip->request, "configuration") == 0)
    {
	html_sysheader(fd, "H2", "Phttpd Configuration");

	fd_puts("<H3>General</H3>\n", fd);

	fd_puts("<PRE>\n", fd);
	
	fd_printf(fd, "debug                %d\n", debug);
	fd_printf(fd, "listen_sock          %d\n", listen_sock);
	fd_printf(fd, "user_dir             %s\n", user_dir ? user_dir : "");
	fd_printf(fd, "server_host          %s\n",
		  server_host ? server_host : "");
	fd_printf(fd, "server_port          %d\n", server_port);
	fd_printf(fd, "server_uid           %d\n", server_uid);
	fd_printf(fd, "server_gid           %d\n", server_gid);
	fd_printf(fd, "n_listen             %d\n", n_listen);
	fd_printf(fd, "n_accept             %d\n", n_accept);
	fd_printf(fd, "n_processes          %d\n", n_processes);
	fd_printf(fd, "concurrency          %d\n", concurrency);
	fd_printf(fd, "gc_time              %d\n", gc_time);
	fd_printf(fd, "gc_sleep             %d\n", gc_sleep);
	fd_printf(fd, "select_timeout       %d\n", select_timeout);
	fd_printf(fd, "max_fds              %d\n", max_fds);
	
	fd_puts("</PRE>\n", fd);
	
	if (url_handlers_table)
	{
	    fd_puts("<H3>URL Handlers</H3>\n", fd);
	    
	    fd_puts("<PRE>\n", fd);
	    tbl_foreach(url_handlers_table, print_table_pair, &fd);
	    fd_puts("</PRE>\n", fd);
	}

	if (url_redirects_table)
	{
	    fd_puts("<H3>URL Redirects</H3>\n", fd);
	    
	    fd_puts("<PRE>\n", fd);
	    tbl_foreach(url_redirects_table, print_table_pair, &fd);
	    fd_puts("</PRE>\n", fd);
	}
    }
    
    else if (strcmp(hip->request, "statistics") == 0)
    {
	struct count_t count;
	char buf[256];
	int cs, dt;
	time_t time_now;

	
	time(&time_now);

	html_sysheader(fd, "H2", "Phttpd Statistics");

	fd_puts("<H3>General</H3>", fd);

	fd_puts("<PRE>", fd);
	
	fd_printf(fd, "\tServer started:\t%s",
		  ctime_r(&start_time, buf, sizeof(buf)));

	fd_printf(fd, "\tCurrent time:\t%s",
		  ctime_r(&time_now, buf, sizeof(buf)));
	
	fd_printf(fd, "\n\tConnections totally:\t%d ",
		  cip->request_no);

	cs = cip->request_no / (time_now - start_time);
	if (cs == 0)
	{
	    dt = (time_now - start_time) / 60 + 1;
	    cs = cip->request_no / dt;
	    
	    if (cs == 0)
	    {
		dt = (time_now - start_time) / 3600 + 1;
		cs = cip->request_no / dt;
		fd_printf(fd, "(%d conns/hour)", cs);
	    }
	    else
		fd_printf(fd, "(%d conns/min)", cs);
	}
	else
	    fd_printf(fd, "(%d conns/sec)", cs);

	fd_printf(fd, "\n\t<A HREF=\"%s?connections\">", hip->url);
	
	fd_printf(fd, "Connections currently</A>:\t%d\n",
		  clist_active());
	fd_puts("</PRE>", fd);

	fd_puts("<H3>Cache statistics</H3>\n", fd);
	
	{
	    cachestat_t csb;
	    
	    fd_puts("<PRE><TABLE WIDTH=90%><TR>", fd);
	    fd_printf(fd, "\t<B><TH ALIGN=LEFT><B>%-8s<TH ALIGN=RIGHT>%10s<TH ALIGN=RIGHT>%10s<TH ALIGN=RIGHT>%10s<TH ALIGN=RIGHT>%10s</B>\n",
		      "Cache",
		      "Lookups",
		      "Reloads",
		      "Updates",
		      "Hits");
		      
	    urlcache_getstats(&csb);
	    print_stats(fd, &csb, "URL");
	    
	    fscache_getstats(&csb);
	    print_stats(fd, &csb, "Filesys");
	    
	    usercache_getstats(&csb);
	    print_stats(fd, &csb, "Username");

	    hostcache_getstats_byname(&csb);
	    print_stats(fd, &csb, "Hostname");
	    
	    hostcache_getstats_byaddr(&csb);
	    print_stats(fd, &csb, "Hostaddr");
	    
	    fd_puts("</TABLE></PRE>\n", fd);
	}

#ifdef INCLUDE_ALLOC_STATS
	{
	    struct allocstats lasb;
	    
	    fd_puts("<H3>Memory allocation statistics</H3>\n", fd);

	    mutex_lock(&alloc_stats.lock);
	    lasb = alloc_stats;
	    mutex_unlock(&alloc_stats.lock);

	    fd_puts("<PRE>", fd);
	    fd_printf(fd, "\tCalls to malloc\t\t%lu\n", lasb.malloc_calls);
	    fd_printf(fd, "\tCalls to free\t\t%lu\n", lasb.free_calls);
	    fd_printf(fd, "\tCalls to realloc\t%lu\n", lasb.realloc_calls);
	    fd_printf(fd, "\tBytes allocated\t\t%lu\n", lasb.bytes);
	    fd_printf(fd, "\tBlocks allocated\t%lu\n", lasb.blocks);
	    fd_puts("</PRE>", fd);
	}
#endif
	
	if (cntfile_path)
	{
	    get_totals(&count);
	    
	    fd_printf(fd,
	"<H4>Totals for <A HREF=\"%s?requests\">Summary</A> Period: %s",
		      hip->url,
		      ctime_r(&(count.starttime), buf, sizeof(buf)));
	    fd_printf(fd, " to %s</H4>\n",
		      ctime_r(&time_now, buf, sizeof(buf)));
	    print_totals(fd);

	}

    }

    else if (cntfile_path && strcmp(hip->request, "requests") == 0)
    {
	html_sysheader(fd, "H2", "Phttpd Requests Summary");

	print_counts(fd);
    }
    
    else if (strcmp(hip->request, "connections") == 0)
    {
	struct miscinfo mib;
	
	html_sysheader(fd, "H2", "Phttpd Connections");

	mib.fd = fd;
	mib.url = hip->url;
	
	fd_puts("<H3>Currently Available</H3>\n", fd);
	fd_puts("<DL>", fd);
	clist_foreach(print_connection, (void *) &mib);
	fd_puts("</DL>", fd);
    }
    
    else if (strcmp(hip->request, "modules") == 0)
    {
	html_sysheader(fd, "H2", "Phttpd Modules");
	
	fd_puts("<H3>Currently Loaded</H3>\n", fd);
	md_foreach(print_module, (void *) &fd);
    }
    
    else if (strcmp(hip->request, "descriptors") == 0)
    {
	html_sysheader(fd, "H2", "Phttpd Descriptors");
	fd_puts("<H3>Currently Opened (excluding this connection)</H3>\n", fd);

	fd_puts("<DL>\n", fd);
	fd_foreach(print_fd, (void *) &fd);
	fd_puts("</DL>\n", fd);
    }

    else
    {
	html_sysheader(fd, "H2", "Unknown config request");
	fd_puts("<H3>Unknown config request:<H3>", fd);

	fd_puts("<BLOCKQUOTE>", fd);
	fd_puts(hip->request, fd);
	fd_puts("</BLOCKQUOTE>", fd);
    }
    
    html_sysfooter(fd);
#if 0
    hip->length = fd_written(fd) - len;
#endif
    return result;
}




int pm_request(struct connectioninfo *cip)
{
    struct httpinfo *hip = cip->hip;

    
    if (strcasecmp(hip->method, "GET") == 0 ||
	strcasecmp(hip->method, "HEAD") == 0)
	return http_get_head(cip);
    else
	return -1;
}
